"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const util_1 = __importDefault(require("util"));
const error_1 = require("./error");
const isnode_1 = __importDefault(require("./isnode"));
let fsprom;
if (isnode_1.default) {
    fsprom = util_1.default.promisify(fs_1.default.readFile);
}
class Las {
    static chunk(arr, size) {
        const overall = [];
        let index = 0;
        while (index < arr.length) {
            overall.push(arr.slice(index, index + size));
            index += size;
        }
        return overall;
    }
    static removeComment(str) {
        return str
            .trim()
            .split('\n')
            .map(val => val.trimLeft())
            .filter(f => !(f.charAt(0) === '#'))
            .join('\n');
    }
    static convertToValue(s) {
        return Boolean(+s) || /^0|0$/.test(s) ? +s : s;
    }
    /**
     * Creates an instance of Las.
     * @param {string} path - Absolute path to las file
     * @memberof Las
     */
    constructor(path) {
        this.path = path;
        this.blobString = this.initialize();
    }
    /**
     * Returns a column in a las file
     * @param {string} column - name of column
     * @returns {(Promise<Array<string| number>>)}
     * @memberof Las
     */
    column(column) {
        return __awaiter(this, void 0, void 0, function* () {
            const hds = yield this.header();
            const sB = yield this.data();
            const index = hds.findIndex(item => item.toLowerCase() === column.toLowerCase());
            if (index < 0) {
                throw new error_1.ColumnError(column);
            }
            return sB.map(c => c[index]);
        });
    }
    /**
     * Returns a column in a las file stripped off null values
     * @param {string} column - name of column
     * @returns {(Promise<Array<string| number>>)}
     * @memberof Las
     */
    columnStripped(column) {
        return __awaiter(this, void 0, void 0, function* () {
            const hds = yield this.header();
            const sB = yield this.dataStripped();
            const index = hds.findIndex(item => item.toLowerCase() === column.toLowerCase());
            if (index >= 0) {
                return sB.map(c => c[index]);
            }
            else {
                throw new error_1.ColumnError(column);
            }
        });
    }
    /**
     * Returns a csv File object in browser | writes csv file to current working driectory in Node
     * @param {string} filename
     * @returns {(Promise<File | void>)}
     * @memberof Las
     */
    toCsv(filename = 'file') {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const headers = yield this.header();
                const data = yield this.data();
                const rHd = headers.join(',') + '\n';
                const rData = data.map(d => d.join(',')).join('\n');
                if (isnode_1.default) {
                    fs_1.default.writeFile(`${filename}.csv`, rHd + rData, 'utf8', err => {
                        if (err) {
                            throw new error_1.CsvError();
                        }
                        console.log(`${filename}.csv has been saved to current working directory`);
                    });
                }
                else {
                    const file = new File([rHd + rData], `${filename}.csv`);
                    return file;
                }
            }
            catch (error) {
                throw new error_1.CsvError();
            }
        });
    }
    /**
     * Returns a csv File object in browser and writes csv file to current working driectory in Node of data stripped of null values
     * @param {string} filename
     * @returns {(Promise<File | void>)}
     * @memberof Las
     */
    toCsvStripped(filename = 'file') {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const headers = yield this.header();
                const data = yield this.dataStripped();
                const rHd = headers.join(',') + '\n';
                const rData = data.map(d => d.join(',')).join('\n');
                if (!isnode_1.default) {
                    const file = new File([rHd + rData], `${filename}.csv`);
                    return file;
                }
                fs_1.default.writeFile(`${filename}.csv`, rHd + rData, 'utf8', err => {
                    if (err) {
                        throw new error_1.CsvError();
                    }
                    console.log(`${filename}.csv has been saved to current working directory`);
                });
            }
            catch (error) {
                throw new error_1.CsvError();
            }
        });
    }
    /**
     * Returns the number of rows in a .las file
     * @returns number
     * @memberof Las
     */
    rowCount() {
        return __awaiter(this, void 0, void 0, function* () {
            const l = yield this.data();
            return l.length;
        });
    }
    /**
     * Returns the number of columns in a .las file
     * @returns number
     * @memberof Las
     */
    columnCount() {
        return __awaiter(this, void 0, void 0, function* () {
            const l = yield this.header();
            return l.length;
        });
    }
    /**
     * Returns a two-dimensional array of data in the log
     * @returns {(Promise<Array<Array<string | number>>>)}
     * @memberof Las
     */
    data() {
        return __awaiter(this, void 0, void 0, function* () {
            const s = yield this.blobString;
            const hds = yield this.header();
            const totalheadersLength = hds.length;
            const sB = s
                .split(/~A(?:\w*\s*)*\n/)[1]
                .trim()
                .split(/\s+/)
                .map(m => Las.convertToValue(m.trim()));
            if (sB.length < 0) {
                throw new error_1.LasError('No data/~A section in the file');
            }
            const con = Las.chunk(sB, totalheadersLength);
            return con;
        });
    }
    /**
     * Returns a two-dimensional array of data in the log with all rows containing null values stripped off
     * @returns {(Promise<Array<Array<string | number>>>)}
     * @memberof Las
     */
    dataStripped() {
        return __awaiter(this, void 0, void 0, function* () {
            const s = yield this.blobString;
            const hds = yield this.header();
            const well = yield this.property('well');
            const nullValue = well.NULL.value;
            const totalheadersLength = hds.length;
            const sB = s
                .split(/~A(?:\w*\s*)*\n/)[1]
                .trim()
                .split(/\s+/)
                .map(m => Las.convertToValue(m.trim()));
            if (sB.length < 0) {
                throw new error_1.LasError('No data/~A section in the file');
            }
            const con = Las.chunk(sB, totalheadersLength);
            const filtered = con.filter(f => !f.some(x => x === +nullValue));
            return filtered;
        });
    }
    /**
     * Returns the version number of the las file
     * @returns {Promise<number>}
     * @memberof Las
     */
    version() {
        return __awaiter(this, void 0, void 0, function* () {
            const v = yield this.metadata();
            return v[0];
        });
    }
    /**
     * Returns true if the las file is of wrapped variant and false otherwise
     * @returns {Promise<boolean>}
     * @memberof Las
     */
    wrap() {
        return __awaiter(this, void 0, void 0, function* () {
            const v = yield this.metadata();
            return v[1];
        });
    }
    /**
     * Returns an extra info about the well stored in ~others section
     * @returns {Promise<string>}
     * @memberof Las
     */
    other() {
        return __awaiter(this, void 0, void 0, function* () {
            const s = yield this.blobString;
            const som = s.split(/~O(?:\w*\s*)*\n\s*/i)[1];
            let str = '';
            if (som) {
                const some = som
                    .split('~')[0]
                    .replace(/\n\s*/g, ' ')
                    .trim();
                str = Las.removeComment(some);
            }
            if (str.length <= 0) {
                throw new error_1.LasError('No ~other section');
            }
            return str;
        });
    }
    /**
     * Returns an array of strings of the logs header/title
     * @returns {Promise<string[]>}
     * @memberof Las
     */
    header() {
        return __awaiter(this, void 0, void 0, function* () {
            const s = yield this.blobString;
            const sth = s.split(/~C(?:\w*\s*)*\n\s*/)[1].split('~')[0];
            const uncommentedSth = Las.removeComment(sth).trim();
            if (uncommentedSth.length < 0) {
                throw new error_1.LasError('There is no header in the file');
            }
            return uncommentedSth.split('\n').map(m => m.trim().split(/\s+|[.]/)[0]);
        });
    }
    /**
     * Returns an object, each well header and description as a key-value pair
     * @returns {Promise<{[key:string]: string}>}
     * @memberof Las
     */
    headerAndDescr() {
        return __awaiter(this, void 0, void 0, function* () {
            const cur = (yield this.property('curve'));
            const hd = Object.keys(cur);
            const descr = Object.values(cur).map((c, i) => (c.description === 'none' ? hd[i] : c.description));
            const obj = {};
            hd.map((_, i) => (obj[hd[i]] = descr[i]));
            if (Object.keys(obj).length < 0) {
                throw new error_1.LasError('Poorly formatted ~curve section in the file');
            }
            return obj;
        });
    }
    /**
     * Returns details of  well parameters.
     * @returns {Promise<WellProps>}
     * @memberof Las
     */
    wellParams() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.property('well');
        });
    }
    /**
     * Returns details of  curve parameters.
     * @returns {Promise<WellProps>}
     * @memberof Las
     */
    curveParams() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.property('curve');
        });
    }
    /**
     * Returns details of  parameters of the well.
     * @returns {Promise<WellProps>}
     * @memberof Las
     */
    logParams() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.property('param');
        });
    }
    metadata() {
        return __awaiter(this, void 0, void 0, function* () {
            const str = yield this.blobString;
            const sB = str
                .trim()
                .split(/~V(?:\w*\s*)*\n\s*/)[1]
                .split(/~/)[0];
            const sw = Las.removeComment(sB);
            const refined = sw
                .split('\n')
                .map(m => m.split(/\s{2,}|\s*:/).slice(0, 2))
                .filter(f => Boolean(f));
            const res = refined.map(r => r[1]);
            const wrap = res[1].toLowerCase() === 'yes' ? true : false;
            if ([+res[0], wrap].length < 0) {
                throw new error_1.LasError("Couldn't get metadata");
            }
            return [+res[0], wrap];
        });
    }
    property(p) {
        return __awaiter(this, void 0, void 0, function* () {
            const regDict = {
                curve: '~C(?:\\w*\\s*)*\\n\\s*',
                param: '~P(?:\\w*\\s*)*\\n\\s*',
                well: '~W(?:\\w*\\s*)*\\n\\s*'
            };
            const regExp = new RegExp(regDict[p], 'i');
            const str = yield this.blobString;
            const substr = str.split(regExp);
            let sw = '';
            if (substr.length > 1) {
                const res = substr[1].split(/~/)[0];
                sw = Las.removeComment(res);
            }
            if (sw.length > 0) {
                const s = {};
                sw.split('\n').map(c => {
                    const obj = c.replace(/\s*[.]\s+/, '   none   ');
                    const title = obj.split(/[.]|\s+/)[0];
                    const unit = obj
                        .trim()
                        .split(/^\w+\s*[.]*s*/)[1]
                        .split(/\s+/)[0];
                    const description = Boolean(obj.split(/[:]/)[1].trim()) ? obj.split(/[:]/)[1].trim() : 'none';
                    const third = obj.split(/[:]/)[0].split(/\s{2,}\w*\s{2,}/);
                    let value = third.length > 2 && !Boolean(third[third.length - 1]) ? third[third.length - 2] : third[third.length - 1];
                    value = value.length > 0 ? value.trim() : value;
                    s[title] = { unit, value, description };
                });
                return s;
            }
            else {
                throw new error_1.PropertyError(p);
            }
        });
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            if (isnode_1.default) {
                try {
                    const str = yield fsprom(this.path, 'utf8');
                    return str;
                }
                catch (error) {
                    throw new error_1.PathError();
                }
            }
            else {
                if (this.path instanceof File) {
                    // eslint-disable-next-line @typescript-eslint/explicit-function-return-type
                    return new Promise((resolve, reject) => {
                        const reader = new FileReader();
                        reader.onerror = () => {
                            reject(new error_1.PathError());
                        };
                        reader.onload = () => {
                            resolve(reader.result);
                        };
                        reader.readAsText(this.path);
                    });
                }
                else {
                    try {
                        const val = yield fetch(this.path);
                        const text = yield val.text();
                        if (text.includes('404: Not Found')) {
                            throw new error_1.PathError();
                        }
                        return text;
                    }
                    catch (error) {
                        throw new error_1.PathError();
                    }
                }
            }
        });
    }
}
exports.Las = Las;
